-- Markov: A markovian text generator
-- Author: Ricardo M. Forno
-- Version 1.1 - 2006/01/27
include get.e
include machine.e

constant MAXINT = 1073741823

function ivalue(sequence x)
-- Converts a string into a non-negative integer
    atom a
    sequence s
    s = value(x)
    if s[1] != GET_SUCCESS then
	return -1
    end if
    a = s[2]
    if not integer(a) then
	return -1
    end if
    if a < 0 then
	return -1
    end if
    return a
end function

function getargs()
-- Reads and processes command line
    sequence cmd
    integer n, r
    cmd = command_line()
    if length(cmd) != 6 then
	puts(2, "Syntax: ex markov <s> <d> <n> <r>\n")
	puts(2, "where:\n")
	puts(2, "<s> is the source file name\n")
	puts(2, "<d> is the destination file name\n")
	puts(2, "<n> is the order of the n-gram applied\n")
	puts(2, "<r> is a random seed; when zero, it is haphazard\n")
	abort(1)
    end if
    n = ivalue(cmd[5])
    if n = -1 then
	puts(2, "Bad Markov Chain order\n")
	abort(2)
    end if
    r = ivalue(cmd[6])
    if r = -1 then
	puts(2, "Bad random seed\n")
	abort(3)
    end if
    return {cmd[3], cmd[4], n, r}
end function

function readdata(sequence filename)
-- Reads source file into RAM
    sequence data
    integer fn, c
    data = {}
    fn = open(filename, "rb")
    if fn = -1 then
	puts(2, "Cannot open file " & filename & " for reading\n")
	abort(4)
    end if
    while 1 do -- Read data bytes up to end of file
	c = getc(fn)
	if c = -1 then
	    return data
	end if
	data &= c
    end while
    close(fn)
end function

function gentab(sequence data, integer n)
-- Generates hash tables
    sequence table, hsh
    integer len, ltab, hash, acum, nm1, dim1, o1, h, ltab1, mult
    object o
    len = length(data)
    nm1 = n - 1
    ltab = len - nm1
    ltab1 = len + 1
    mult = floor(MAXINT / ltab1)
    hsh = repeat(0, ltab)
    table = repeat(0, ltab1)
    hash = 0
    acum = 0
    for i = 1 to n do -- Get hash for 1rst element only
	acum += data[i]
	hash += acum
	hash = remainder(hash, ltab1) -- Here hash is always >= 0
    end for
    h = remainder(hash * mult, ltab1) + 1
    hsh[1] = h
    table[h] = {1}
    for i = 2 to ltab - 1 do
	dim1 = data[i - 1]
	acum += data[i + nm1] - dim1
	hash += acum - n * dim1
	hash = hash - ltab1 * floor(hash / ltab1) -- Inclusive for hash < 0)
	h = hash * mult
	h = 1 + remainder(h, ltab1) 
	while 1 do
	    o = table[h]
	    if atom(o) then
		table[h] = {i}
		exit
	    else
		o1 = o[1]
		if equal(data[o1..o1 + nm1], data[i..i + nm1]) then
		    table[h] &= i
		    exit
		else
		    h += 1
		    if h > ltab1 then
			h = 1
		    end if
		end if
	    end if
	end while
	hsh[i] = h
    end for
    return {hsh, table}
end function

procedure gentext(integer fn, sequence data, sequence hsh, sequence table,
	integer n)
-- Generates markovian text when n > 0
    integer k, h, len, r, lent
    sequence t
    len = length(data)
    puts(fn, data[1..n])
    k = 1
    while 1 do
	h = hsh[k]
	if h = 0 then
	    exit
	end if
	t = table[h]
	lent = length(t)
	if lent = 1 then
	    k = t[1]
	else
	    while 1 do -- Avoid repetition
		r = rand(lent)
		if t[r] != k then
		    exit
		end if
	    end while
	    k = t[r]
	end if
	puts(fn, data[k + n])
	k += 1
    end while
end procedure

procedure gentext0(integer fn, sequence data)
-- Generates markovian text when n = 0, for better speed
    integer len, r, r1
    len = length(data)
    r1 = 1
    puts(fn, data[r1])
    while 1 do
	while 1 do
	    r = rand(len)
	    if r != r1 then -- Avoid repetition
		exit
	    end if
	    if len = 1 then -- Repetition unavoidable in this case
		exit
	    end if
	end while
	r1 = r
	puts(fn, data[r])
	if r = len then
	    exit
	end if
    end while
end procedure

procedure markov()
-- Generates markovian text
    sequence data, cmd, hsh, table
    integer len, n, r, fn
    cmd = getargs()
    data = readdata(cmd[1])
    len = length(data)
    if len = 0 then
	puts(2, "Data length must not be 0\n")
	abort(5)
    end if
    n = cmd[3]
    r = cmd[4]
    fn = open(cmd[2], "wb")
    if fn = -1 then
	puts(2, "Cannot open file " & cmd[2] & " for writing\n")
	abort(6)
    end if
    if r then
	set_rand(r)
    end if
    if len < n then
	puts(2, "Data length should not be less than Markov order\n")
	abort(7)
    elsif n = 0 then -- Speed up for the simplest case
	gentext0(fn, data)
    else -- Normal cases, when n > 0
	table = gentab(data, n)
	hsh = table[1]
	table = table[2]
	gentext(fn, data, hsh, table, n)
    end if
    close(fn)
end procedure

markov()
